extern crate rltk;
use rltk::{Point, Rltk, VirtualKeyCode};
extern crate specs;
use super::{CombatStats, Map, Position, MoveEvent, RunState, State, Viewshed, WantsToMelee, InputEvent};
use specs::prelude::*;


fn clamp<T: PartialOrd>(low: T, high: T, val: T) -> T {
    if val < low {
        low
    } else if val > high {
        high
    } else {
        val
    }
}

pub struct HandleMoveEvent;

impl<'a> System<'a> for HandleMoveEvent {
    type SystemData = (
        WriteStorage<'a, MoveEvent>,
        WriteStorage<'a, Position>,
        WriteStorage<'a, Viewshed>,
        WriteExpect<'a, Point>,
        ReadExpect<'a, Map>,
    );

    fn run(&mut self, (mut move_events, mut pos, mut viewshed, mut point, map): Self::SystemData) {
        for (&event, pos, viewshed) in (&move_events, &mut pos, &mut viewshed).join() {
            let MoveEvent { destination, tgt_x, tgt_y } = event;
            if !map.blocked[destination] {
                viewshed.dirty = true;
                pos.x = tgt_x;
                pos.y = tgt_y;
                point.x = clamp(0, 79, tgt_x);
                point.y = clamp(0, 49, tgt_y);
            }
        }
        move_events.clear();
    }
}


pub struct HandleInputEvent;

impl<'a> System<'a> for HandleInputEvent {
    type SystemData = (
        WriteStorage<'a, InputEvent>,
        ReadExpect<'a, Map>,
        Entities<'a>,
        ReadStorage<'a, Position>,
        ReadStorage<'a, CombatStats>,
        WriteStorage<'a, WantsToMelee>,
        WriteStorage<'a, MoveEvent>,
    );

    fn run(&mut self, (mut event, map, entities, position, stats, mut attack, mut mov): Self::SystemData) {
        for (e, ev, pos) in (&entities, &event, &position).join() {
            match *ev {
                InputEvent::PlayerMovement { delta_x, delta_y } => {
                    // depending on what's in the target cell, we
                    // either want to attack or move
                    let tgt_x = pos.x + delta_x;
                    let tgt_y = pos.y + delta_y;
                    let destination = map.xy_idx(tgt_x, tgt_y);
                    if let Some(target) = map.get_attackable(destination, &stats) {
                        attack.insert(e, WantsToMelee { target })
                            .expect("Unable to add WantsToAttack");
                    } else {
                        mov.insert(e, MoveEvent { destination, tgt_x, tgt_y })
                            .expect("Unable to add MoveEvent");
                    }
                }
            }
        }
        event.clear();
    }
}

pub fn player_input(gs: &mut State, ctx: &mut Rltk) -> RunState {
    let player = *gs.ecs.fetch::<Entity>();
    let mut input = gs.ecs.write_storage::<InputEvent>();
    // Player movement
    match ctx.key {
        None => return RunState::AwaitingInput, // Nothing happened
        Some(key) => match key {
            VirtualKeyCode::Left | VirtualKeyCode::Numpad4 | VirtualKeyCode::H =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: -1, delta_y: 0 }),

            VirtualKeyCode::Right | VirtualKeyCode::Numpad6 | VirtualKeyCode::L =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 1, delta_y: 0 }),

            VirtualKeyCode::Up | VirtualKeyCode::Numpad8 | VirtualKeyCode::K =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 0, delta_y: -1 }),

            VirtualKeyCode::Down | VirtualKeyCode::Numpad2 | VirtualKeyCode::J =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 0, delta_y: 1 }),

            // Diagonals
            VirtualKeyCode::Numpad9 | VirtualKeyCode::U =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 1, delta_y: -1 }),

            VirtualKeyCode::Numpad7 | VirtualKeyCode::Y =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: -1, delta_y: -1 }),

            VirtualKeyCode::Numpad3 | VirtualKeyCode::N =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: 1, delta_y: 1 }),

            VirtualKeyCode::Numpad1 | VirtualKeyCode::B =>
                input.insert(player, InputEvent::PlayerMovement { delta_x: -1, delta_y: 1 }),

            _ => return RunState::AwaitingInput,
        },
    }.expect("Unable to insert player movement");
    RunState::PlayerTurn
}
